from packtivity.asyncbackends import ForegroundProxy, ForegroundBackend
import logging

log = logging.getLogger(__name__)

class InitProxy(ForegroundProxy):
    def __init__(self, *args, **kwargs):
        super(InitProxy,self).__init__(*args, **kwargs)
    def proxyname(self):
        return 'InitProxy'

class InitBackend(ForegroundBackend):
    def submit(self, spec, parameters, state, metadata):
        foreground_proxy = super(InitBackend,self).submit(spec, parameters, state,metadata)
        return InitProxy(foreground_proxy.result, foreground_proxy.success)



import networkx as nx

def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None

def reset_step(workflow, step):
    s = workflow.dag.getNode(step)
    reset_node_state(s)
    try:
        s.task.state.reset()
    except AttributeError:
        pass

def reset_steps(workflow, steps):
    for s in steps:
        reset_step(workflow, s)

def collective_downstream(workflow, steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag, step):
            downstream.add(x)
    return list(downstream)




import yadageschemas

def workflow(source, toplevel, schema_name='yadage/workflow-schema', schemadir=None, validate=True):
    '''
    load (and validate) workflow from source
    param source: URI fragment to the source
    param toplevel: base URI to resolve references from
    schemadir
    '''
    data = yadageschemas.load(source, toplevel, schema_name, schemadir, validate)
    return data

def validate(data, schema_name='yadage/workflow-schema', schemadir=None):
    '''validate workflow data'''
    yadageschemas.validator(schema_name,schemadir).validate(data)



#!/usr/bin/env python
import logging
import click
import os
from packtivity.statecontexts.posixfs_context import LocalFSState
from yadage.state_providers.localposix import LocalFSProvider

from .steering_object import YadageSteering
from .controllers import PersistentController
from .wflowstate import load_model_fromstring
from .stages import JsonStage
import yadage.manualutils as manualutils
import yadage.utils as utils
import yadage.interactive as interactive
import yadage.reset as reset_module
import yadage.workflow_loader as workflow_loader

log = logging.getLogger(__name__)
logging.basicConfig(level = logging.INFO)

@click.group()
def mancli():
    pass

@mancli.command()
@click.option('-s', '--statesetup', default='filebacked:yadage_state.json')
@click.option('-t', '--toplevel', default=os.getcwd())
@click.option('-d', '--dataopt', multiple=True, default=None, help = 'options for the workflow data state')
@click.option('--metadir', default=None, help = 'directory to store workflow metadata')
@click.option('--parameter', '-p', multiple=True)
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', nargs = -1)
def init(workdir, workflow, initfiles, statesetup, dataopt, metadir, toplevel, parameter):
    initdata = utils.getinit_data(initfiles, parameter)
    dataopts = utils.options_from_eqdelimstring(dataopt)

    ys = YadageSteering()
    ys.prepare(workdir,
        metadir = metadir,
        dataopts = dataopts
    )
    ys.init_workflow(workflow, initdata = initdata, toplevel = toplevel, modelsetup = statesetup)


def click_print_applicable_stages(controller):
    click.secho('Applicable Stages: ', fg='blue')
    for x in controller.adageobj.rules:
        if x.identifier in controller.applicable_rules():
            click.secho('{}/{}'.format(x.offset, x.rule.name))

def click_print_submittable_nodes(controller):
    click.secho('Submittable Nodes: ', fg='blue')
    _, s2r = manualutils.rule_steps_indices(controller.adageobj)
    for x in controller.adageobj.dag.nodes():
        node = controller.adageobj.dag.getNode(x)
        rule = controller.adageobj.view().getRule(identifier = s2r[node.identifier])
        if node.identifier in controller.submittable_nodes():
            click.secho('node: {} ({}) part of stage {}'.format(node.name, node.identifier,  '/'.join([rule.offset,rule.rule.name])))

@mancli.command()
@click.option('-n','--name', default=None)
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-v', '--verbosity', default='ERROR')
@click.option('-b', '--backend', default='celery')
def apply(name, statetype, verbosity, backend,modelopt):
    logging.basicConfig(level=getattr(logging, verbosity))

    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    backend =  utils.setupbackend_fromstring(backend)
    controller = PersistentController(model,backend)

    if not name:
        click_print_applicable_stages(controller)
        return

    offset, name = name.rsplit('/',1)
    rule = controller.adageobj.view(offset).getRule(name)
    if not rule:
        click.secho('No such stage, pick one of the applicable below:', fg='red')
        click_print_applicable_stages(controller)
        return

    if rule in controller.adageobj.applied_rules:
        click.secho('Already applied.', fg = 'yellow')
        return

    if rule.identifier not in controller.applicable_rules():
        click.secho('Rule not yet appilcable', fg = 'red')
        return

    controller.apply_rules([rule.identifier])
    click.secho('stage applied', fg = 'green')

@mancli.command()
@click.option('-n','--nodeid', default=None)
@click.option('-a','--allof', default=None)
@click.option('-o', '--offset', default='')
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-v', '--verbosity', default='ERROR')
@click.option('-b', '--backend', default='celery')
def submit(nodeid, allof, offset, statetype, modelopt, verbosity, backend):
    logging.basicConfig(level=getattr(logging, verbosity))

    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    backend =  utils.setupbackend_fromstring(backend)
    controller = PersistentController(model,backend)

    if not (allof or nodeid):
        click_print_submittable_nodes(controller)
        return

    if nodeid:
        nodes_to_submit = [nodeid] if nodeid in controller.submittable_nodes() else []
    if allof:
        offset, name = allof.rsplit('/',1)
        rule = controller.adageobj.view().getRule(name = name, offset = offset)
        if not rule:
            click.secho('stage not found!', fg = 'red')
            return


        all_submittable = controller.submittable_nodes()
        _, s2r = manualutils.rule_steps_indices(controller.adageobj)
        nodes_to_submit = [x for x in all_submittable if s2r[x] == rule.identifier]

    if not nodes_to_submit:
        click.secho('No nodes to submit (perhaps already submitted?)', fg = 'yellow')
        return

    controller.submit_nodes(nodes_to_submit)
    click.secho('submitted: {}'.format(nodes_to_submit), fg = 'green')

@mancli.command()
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-b', '--backend', default='celery')
def show(statetype, backend, modelopt):
    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    controller = PersistentController(model)
    controller.backend = utils.setupbackend_fromstring(backend)

    click.secho('''
Workflow:
---------
state source: {statetype}
successful: {successful}
finished: {finished}
valid: {valid}
# of applicable rules: {applicable}
# of submittable nodes: {submittable}
'''.format(
    statetype = statetype,
    successful =  click.style(str(controller.successful()), fg = 'green' if controller.successful() else 'red'),
    finished = click.style(str(controller.finished()), fg = 'green' if controller.finished() else 'yellow'),
    valid = click.style(str(controller.validate()), fg = 'green' if controller.validate() else 'red'),
    applicable = len(controller.applicable_rules()),
    submittable = len(controller.submittable_nodes())
    ))
    click_print_applicable_stages(controller)
    click_print_submittable_nodes(controller)


@mancli.command()
@click.argument('name')
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-b', '--backend', default='celery')
def preview(name,statetype,backend,modelopt):
    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    controller = PersistentController(model)
    controller.backend = utils.setupbackend_fromstring(backend)

    new_rules, new_nodes = manualutils.preview_rule(controller.adageobj, name)
    click.secho('Preview of Stage: # new rules: {} # new nodes {}'.format(
        len(new_rules), len(new_nodes)))
    for n in new_nodes:
        click.secho(
            '-> new node "{}" with {} upstream dependencies'.format(n['name'], len(n['parents'])))
@mancli.command()
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-v', '--verbosity', default='ERROR')
@click.option('-n', '--nsteps', default=-1, help = 'number of steps to process. use -1 to for no limit (will run workflow to completion)')
@click.option('-u', '--update-interval', default=1)
@click.option('-b', '--backend', default='celery')
def step(statetype, verbosity, nsteps, update_interval,backend,modelopt):
    logging.basicConfig(level=getattr(logging, verbosity))

    maxsteps = nsteps if nsteps >= 0 else None
    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    backend = utils.setupbackend_fromstring(backend)

    extend, submit = interactive.interactive_deciders(idbased = True)
    ys = YadageSteering()
    ys.controller = PersistentController(model)
    ys.run_adage(backend,
        maxsteps = maxsteps,
        default_trackers = False,
        submit_decider = submit,
        extend_decider = extend,
        update_interval = update_interval
    )

@mancli.command()
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-v', '--verbosity', default='ERROR')
@click.option('-o', '--offset')
@click.option('-t', '--toplevel', default = os.curdir)
@click.argument('workdir')
@click.argument('workflow')
def add(statetype, verbosity, offset, toplevel, workdir, workflow,modelopt):
    logging.basicConfig(level=getattr(logging, verbosity))

    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    controller = PersistentController(model)

    workflow_json = workflow_loader.workflow(
        workflow,
        toplevel=toplevel,
        validate=True
    )

    state_provider = LocalFSProvider(LocalFSState([workdir]))
    rules = [JsonStage(json, state_provider) for json in workflow_json['stages']]
    with controller.transaction():
        controller.adageobj.view().addWorkflow(rules)

@mancli.command()
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
@click.option('-f', '--fileformat', default='pdf')
@click.option('-w', '--workdir', default=os.curdir)
def visualize(statetype, workdir, fileformat,modelopt):
    from .visualize import write_prov_graph

    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    controller = PersistentController(model)

    write_prov_graph(workdir, controller.adageobj, fileformat)



@mancli.command()
@click.argument('name')
@click.option('-s', '--statetype', default='filebacked:yadage_state.json')
@click.option('-l', '--modelopt', multiple=True, default=None, help = 'options for the workflow state models')
def reset(statetype, name,modelopt):
    stateopts = utils.options_from_eqdelimstring(modelopt)
    model   = load_model_fromstring(statetype,stateopts)
    controller = PersistentController(model)

    offset, name = name.rsplit('/',1)
    rule = controller.adageobj.view(offset).getRule(name)
    if not rule:
        click.secho('state not found!', fg = 'red')

    r2s, _ = manualutils.rule_steps_indices(controller.adageobj)
    steps_of_rule = r2s[rule.identifier]

    to_reset = steps_of_rule + reset_module.collective_downstream(controller.adageobj, steps_of_rule)

    controller.reset_nodes(to_reset)

if __name__ == '__main__':
    mancli()



from contextlib import contextmanager
import logging
import importlib
import yadageschemas
import os

from .steering_object import YadageSteering
from .utils import setupbackend_fromstring
from .interactive import interactive_deciders

log = logging.getLogger(__name__)

def run_workflow(*args, **kwargs):
    '''
    convenience function around steering context, when no additional settings
    are desired.
    '''
    with steering_ctx(*args, **kwargs):
        pass

@contextmanager
def steering_ctx(
    dataarg,
    workflow = None,
    initdata = None,
    toplevel = os.getcwd(),
    backend = None,
    workflow_json = None,
    cache = None,
    dataopts = None,
    updateinterval = 0.02,
    loginterval = 30,
    schemadir = yadageschemas.schemadir,
    metadir = None,
    interactive=False,
    validate=True,
    visualize=True,
    accept_metadir = False,
    modelsetup = 'inmem',
    modelopts = None):
    '''
    context manage around yadage steering object.

    param dataarg: primary data argument (e.g. workdir) of the workflow
    param workflow: workflow spec source
    '''

    ys = YadageSteering()

    if cache:
        accept_metadir = True

    ys.prepare(
        dataarg = dataarg, dataopts = dataopts,
        metadir = metadir, accept_metadir = accept_metadir,
    )
    
    wflow_kwargs = dict() #if this stays empty, error will be raise by ys
    if workflow_json:
        wflow_kwargs = dict(workflow_json = workflow_json)
    elif workflow:
        wflow_kwargs = dict(workflow = workflow, toplevel = toplevel, validate = validate, schemadir = schemadir)

    ys.init_workflow(
        initdata = initdata,
        modelsetup = modelsetup,
        modelopts = modelopts,
        **wflow_kwargs
    )

    custom_tracker = os.environ.get('YADAGE_CUSTOM_TRACKER',None)
    if custom_tracker:
        modulename,trackerclassname = custom_tracker.split(':')
        module = importlib.import_module(modulename)
        trackerclass = getattr(module,trackerclassname)
        ys.adage_argument(additional_trackers = [trackerclass()])

    ys.adage_argument(
        default_trackers = visualize,
        trackevery = loginterval,
        update_interval = updateinterval,
    )
    if interactive:
        extend, submit = interactive_deciders()
        ys.adage_argument(
            extend_decider = extend,
            submit_decider = submit
        )

    yield ys

    backend = backend or setupbackend_fromstring('multiproc:auto')
    log.info('running yadage workflow %s on backend %s', workflow, backend)
    if cache:
        if cache == 'checksums':
            backend.enable_cache(':'.join([cache,os.path.join(ys.metadir,'cache.json')]))
        else:
            backend.enable_cache(cache)

    try:
        ys.run_adage(backend)
    finally:
        ys.serialize()
    if visualize:
        ys.visualize()



import packtivity
import logging
from .utils import leaf_iterator_jsonlike, outputReference

log = logging.getLogger(__name__)

class TaskBase(object):
    def __init__(self, **metadata):
        self.metadata = metadata
        self.inputs = []
        self.parameters = {}
        self.prepublished = None

    def used_input(self, reference):
        self.inputs += [reference]

    def used_inputs(self, references):
        for r in references: self.used_input(r)

    #(de-)serialization
    def json(self):
        return {
            'metadata': self.metadata,
            'parameters': self.parameters,
            'prepublished': self.prepublished,
            'inputs': [x.json() for x in self.inputs]
        }

class packtivity_task(TaskBase):
    '''
    packtivity task
    '''
    def __init__(self, name, spec, state):
        super(packtivity_task, self).__init__(name = name)
        self.spec = spec
        self.state = state

    def s(self, **parameters):
        self.parameters.update(**parameters)
        if self.state:
            for leaf_pointer, leaf_value in leaf_iterator_jsonlike(self.parameters):
                leaf_pointer.set(self.parameters,self.state.contextualize_data(leaf_value))
        # attempt to prepublish output data merely from inputs
        # will still be None if not possible
        self.prepublished = packtivity.prepublish_default(self.spec, self.parameters, self.state)
        log.debug('parameters for packtivity_task set to %s. prepublished result, if any: %s',
                    self.parameters,
                    self.prepublished
        )
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data, state_deserializer):
        instance = cls(data['metadata']['name'], data['spec'], state_deserializer(data['state']) if data['state'] else None)
        instance.parameters = data['parameters']
        instance.prepublished = data['prepublished']
        instance.inputs = map(outputReference.fromJSON, data['inputs'])
        instance.metadata.update(**data['metadata'])
        return instance

    def json(self):
        data = super(packtivity_task, self).json()
        data.update(
            type='packtivity_task',
            spec=self.spec,
            state=self.state.json() if self.state else None,
        )
        return data



import jsonpointer
from .wflowstate import make_deserializer

def rule_steps_indices(workflow):
    rule_to_steps_index = {}
    steps_to_rule_index = {}
    for rule in workflow.rules + workflow.applied_rules:
        path = '/'.join([rule.offset, rule.rule.name])
        p = jsonpointer.JsonPointer(path)
        try:
            steps_of_rule = [x['_nodeid'] for x in p.resolve(workflow.stepsbystage) if '_nodeid' in x]
        except jsonpointer.JsonPointerException:
            steps_of_rule = []
        rule_to_steps_index[rule.identifier] = steps_of_rule
        for step in steps_of_rule:
            steps_to_rule_index[step] = rule.identifier
    return rule_to_steps_index, steps_to_rule_index

def preview_rule(wflow, name = None, identifier=None):
    stateopts = {}
    wflowmaker = make_deserializer(stateopts)
    newflow = wflowmaker(wflow.json())

    if identifier:
        rule = newflow.view().getRule(identifier=identifier)
    else:
        offset, name = name.split('/')
        rule = newflow.view(offset).getRule(name)

    if not rule.applicable(newflow):
        return

    rule.apply(newflow)
    newflow.rules.remove(rule)
    newflow.applied_rules.append(rule)

    existing_rules = [x.identifier for x in (wflow.rules + wflow.applied_rules)]
    existing_nodes = wflow.dag.nodes()

    new_rules = [{'name': x.rule.name, 'offset': x.offset} for x in newflow.rules if x.identifier not in existing_rules]
    new_nodes = [{'name': newflow.dag.getNode(n).name, 'parents': newflow.dag.predecessors(n)} for n in newflow.dag.nodes() if n not in existing_nodes]
    return new_rules, new_nodes



import logging
import contextlib

from adage.wflowcontroller import BaseController
from yadage.wflowstate import load_model_fromstring
from yadage.reset import reset_steps

log = logging.getLogger(__name__)

def setup_controller_from_modelstring(workflowobj = None, modelsetup = 'inmem', modelopts = None):
    '''
    return controller instance based on state configuration. For
    transaction-based states, returns PersistentController, for in-
    memory states returns in BaseController
    '''
    modelopts = modelopts or {}
    if modelsetup == 'inmem':
        return BaseController(workflowobj)
    else:
        model = load_model_fromstring(modelsetup,modelopts,workflowobj)
        return PersistentController(model)

class PersistentController(BaseController):
    '''
    workflow controller, that explicltly calls transaction methods on non read-only operations on the workflow state
    '''
    def __init__(self, model, backend = None):
        '''
        :param model: the model on whih the controller will operate
        :param backend: the backend to against which to check workflow state.

        :return: the controller instance
        '''
        self.model = model
        super(PersistentController, self).__init__(self.model.load(),backend)


    @contextlib.contextmanager
    def transaction(self):
        '''the transaction context. will commit model to persistent store on exit.'''
        self.adageobj = self.model.load()
        yield

        isvalid = self.validate()
        if not isvalid:
            #raise RuntimeError('was about to commit invalid data!')
            log.warning('commit is in valid %s', isvalid)
        self.model.commit(self.adageobj)

    def submit_nodes(self, nodeids):
        '''
        submit nodes to backend

        :param nodeids: list of ids of nodes to be submitted
        :return: None
        '''
        with self.transaction():
            nodes = [self.adageobj.dag.getNode(nodeid) for nodeid in nodeids]
            # log.info('submitting nodes to backend: %s', nodes)
            super(PersistentController,self).submit_nodes(nodes)

    def apply_rules(self, ruleids):
        '''
        apply rules to workflow instance

        :param ruleids: list of ids of reuls to be applied
        :return: None
        '''
        with self.transaction():
            rules = [r for r in self.adageobj.rules if r.identifier in ruleids]
            super(PersistentController,self).apply_rules(rules)

    def sync_backend(self):
        '''
        synchronize node data with backend
        '''
        with self.transaction():
            super(PersistentController,self).sync_backend()

    def applicable_rules(self):
        '''
        :return: a list of rules whose predicate is fulfilled
        '''
        applicable_rules = super(PersistentController,self).applicable_rules()
        return [x.identifier for x in applicable_rules]

    def submittable_nodes(self):
        '''
        :return: a list of nodes with sucessfull and completed upstream
        '''
        submittable_nodes = super(PersistentController,self).submittable_nodes()
        return [x.identifier for x in submittable_nodes]

    def reset_nodes(self, nodeids):
        '''
        :param nodeids: list of ids of nodes to reset
        :return: None
        '''
        with self.transaction():
            reset_steps(self.adageobj,nodeids)
